#!/usr/bin/env python
# Four spaces as indentation [no tabs]
import os
import json
import subprocess
import re
import csv
from pathlib import Path
import sys
import pprint
import time
import random


from Planners.EPDDL.parser import EPDDL_Parser
from Planners.CaseBasedS1 import caseBased_s1_solver
from Planners.CaseBasedS1 import getStates
from Planners.CaseBasedS1 import caseBased_s1_distance
from Planners.PlansformerS1 import plansformer_s1


# Constants
systemALL = -1;
systemONE = 1
systemTWO = 2

plannerALL = -1;

plannerS1_Dist1 = 1
plannerS1_Dist2 = 2
plannerS1_Random = 3
plannerS1_Combined = 4
plannerS1_Plansformer = 5

plannerS2_1 = 1
plannerS2_2 = 2

maxGroundDepth = 2
output_folder = "Output/"
output_folderPl1 = output_folder + "Pl1/"
output_folderEFP = output_folder + "EFP/"
output_folderPDKB = output_folder + "PDKB/"
scripts_folder = "Scripts/"
dbFolder = "Memory/"
#db_file = "memory.db"
jsonFilename = "cases.json"


# Thresholds
threshold1 = 20 #This threshold represents the minimun number of plan (existence) that needs to be generated from S2 (in the same domain) before accepting S1 solution
threshold2 = 0.8 #The value which current_reward/average_revard must surpass to eploy S1 -- FOR NOW NOT EMPLOYED
threshold3 = 0.9 #This value represents the risk-adversion of the system -- The higher it is the more incline to use S2 the system is
threshold4 = 20 #This is the threshold that is used to set 'M' = 1 when S1 hasn't enough experience,
epsilonS1 = 0.1

maxExperience = 1000

# Global variables seen by all the functions
domainFile = ""
problemFile = ""
context = ""
instanceNameEFP = ""
solutionS1 = []
confidenceS1 = 0
totalTIME = time.time()
difficulty = 0



# Global Objects
parser = EPDDL_Parser()

def createFolders():
    Path(dbFolder).mkdir(parents=True,exist_ok=True)
    Path(output_folder).mkdir(parents=True,exist_ok=True)
    Path(output_folderPl1).mkdir(parents=True,exist_ok=True)
    Path(output_folderEFP).mkdir(parents=True,exist_ok=True)
    Path(output_folderPDKB).mkdir(parents=True,exist_ok=True)

def getVarFromFile(filename,varname):
    with open(filename) as myfile:
        for line in myfile:
            if ":"+varname in line:
                line = re.sub(r';.*$', '', line)
                discard, var = line.partition(":"+varname)[::2]
                var = var.replace(")", "")
                myfile.close()
                return var.strip()
        raise Exception('Missing variable named '+ varname)
    raise Exception('Missing file named '+ filename)

def readThreshold(thresholdFile):
    threshold1 = float(getVarFromFile(thresholdFile,"threshold1"))
    threshold2 = float(getVarFromFile(thresholdFile,"threshold2"))
    threshold3 = float(getVarFromFile(thresholdFile,"threshold3"))
    threshold4 = float(getVarFromFile(thresholdFile,"threshold4"))
    epsilonS1  = float(getVarFromFile(thresholdFile,"epsilonS1"))

#Every Solver stores the solution on a tmp file
def readSolutionFromFile(filename):
    try:
        with open(filename) as myfile:
            for line in myfile:
                if "Solution" in line:
                    if '=' in line:
                        discard, sol = line.partition("=")[::2]
                        res=[]
                        sol = sol.replace(' ','')
                        sol = sol.replace(';','')
                        sol = sol.replace('\n','')

                        acts = sol.split(',')

                        for act in acts:
                            res.append(act)

                        myfile.close()
                        return res

        #raise Exception('Missing solution in '+ filename)
        return "noSolution"
    except IOError:
        return "noSolution"

def readTimeFromFile(filename):
    with open(filename) as myfile:
        for line in myfile:
            if "TIMED-OUT" in line:
                return "TO"
            elif "completed the search" in line:
                if ':' in line:
                    discard, ret = line.partition(":")[::2]
                    ret = ret.replace(' ','')
                    ret = ret.replace('\n','')

                    myfile.close()
                    return ret

    #raise Exception('Missing plan in '+ filename)
    return "TO"

def executeS1():

    #New S1###################
    similarity_threshold = 0.2 # If < than this is not returned
    json_path = dbFolder+jsonFilename
    try:
        with open(json_path) as f:
            experience = json.load(f)

        cases = experience["cases"]

        if (plannerS1 == plannerS1_Dist1 or plannerS1 == plannerS1_Dist2):
            init,goal = getStates.States(problemFile) #reading initial and goal states from problem file

            #returned_list has records of this form <path_to_sol, similarity_score, problem_name>
            sol = ""
            confidence = 0
            returned_list = caseBased_s1_distance.doCaseMatch(cases, plannerS1, parser.domain_name, init, goal, similarity_threshold)
            if returned_list:
                first_act = True
                for act in returned_list[0][1]:
                    if first_act:
                        sol = act
                        first_act = False
                    else:
                        sol = sol + ", " + act
                confidence = returned_list[0][0]



        elif(plannerS1 == plannerS1_Combined):
            init,goal = getStates.States(problemFile) #reading initial and goal states from problem file

            #returned_list has records of this form <path_to_sol, similarity_score, problem_name>
            sol = ""
            confidence = 0
            returned_list0 = caseBased_s1_distance.doCaseMatch(cases, 0, parser.domain_name, init, goal, similarity_threshold)
            returned_list1 = caseBased_s1_distance.doCaseMatch(cases, 1, parser.domain_name, init, goal, similarity_threshold)

            if returned_list0 and returned_list1:
                if (returned_list0[0][0] < returned_list1[0][0]):
                    returned_list = returned_list1
                else:
                    returned_list = returned_list0

                first_act = True
                for act in returned_list[0][1]:
                    if first_act:
                        sol = act
                        first_act = False
                    else:
                        sol = sol + ", " + act
                confidence = returned_list[0][0]

        elif (plannerS1 == plannerS1_Random):
            sol = ""
            confidence, plan = caseBased_s1_solver.randomSolve(cases)
            first_act = True
            for act in plan:
                if first_act:
                    sol = act
                    first_act = False
                else:
                    sol = sol + ", " + act

        elif (plannerS1 == plannerS1_Plansformer):
            tens_confidence, plan = plansformer_s1.solve(domainFile,problemFile)
            str_confidence= re.sub(r'tensor\((.+)\)', r'\1', str(tens_confidence))
            confidence = float(str_confidence)
            sol = ""
            first_act = True
            for act in plan:
                if first_act:
                    sol = act
                    first_act = False
                else:
                    sol = sol + ", " + act
        else:
            raise Exception("The requested System 1 has not been implemented yet.")
    #    solString = s1_planner.s1Solver(parser.domain_name,parser.problem_name,json_path)
    #    solString = solString.replace(";", ",")
        resFile = instanceNameEFP.replace(".tmp", "S1.tmp")
        out = open(output_folderPl1 + resFile, "w")
        out.write("Solution = " + sol)
        out.close()
        return confidence, resFile;
    except IOError:
        return 0, "";

def solveWithS1():
    confidence, resFile = executeS1()
    solutionS1 = readSolutionFromFile(output_folderPl1 + resFile)
    return solutionS1, confidence

def validateSolution(solution):
    stringSolution = ""
    for elem in solution:
        stringSolution += " "
        stringSolution += elem
    #print("Execution Line is:  sh ./Planners/EFP/scripts/validate_solution.sh " + instanceNameEFP + " " + stringSolution)
    result = subprocess.run(['sh','./'+ scripts_folder + 'validate_solution.sh', output_folderEFP + instanceNameEFP, stringSolution], stdout=subprocess.PIPE)
    res = result.stdout.decode('utf-8')
    if ("Goal found" in res):
        #print("Found result is " + stringSolution)
        return 1.0
    elif ("The solution score is: <<" in res):
        res = re.findall(r'<<(.+)>>', res, re.MULTILINE)
        return float(res[-1])

    else:
        #print("Goal not found")
        return 0.0

def generateEFPInstance():
    global instanceNameEFP
    global parser

    parser.parse_domain(domainFile)
    parser.parse_problem(problemFile)
    instanceNameEFP = parser.print_EFP(output_folderEFP)

def estimateDifficulty(instanceDepth):
    #For now difficulty evaluation that does not consider goal or initial state (Maybe include planning grpah lenght?)
    #parser.ag_number
    #parser.act_number
    #parser.fl_number
    return (pow(parser.ag_number, instanceDepth)*pow(2, parser.fl_number))

def estimateTimeCons(planner,difficulty):
    dbFilename = dbFolder + jsonFilename
    range = 100
    timeSTARTConsumption = 0
    matchCount = 0
    #Controlla IL CSV e leggi valori
    try:
        with open(dbFilename) as db:
            db_reader = csv.reader(db, delimiter=',')
            for row in db_reader:
                if(row[0] == parser.domain_name):
                    if(int(row[1]) == planner):
                        rowDiff = float(row[2])
                        if ((rowDiff < (difficulty + range)) and (rowDiff > (difficulty - range))):
                            matchCount += 1
                            timeSTARTConsumption += float(row[3])

    except IOError:
        matchCount = 0


    if matchCount > 0:
        return (timeSTARTConsumption/matchCount)


    else:
        noMatchConsumption = 5
        return noMatchConsumption

def checkDCK():
    return (parser.dynamicCK.lower() == "true")

def solveWithS2NoPlan(timeLimit):
    solveWithS2(timeLimit, selectPlannerS2())

def solveWithS2(timeLimit, planner):

    planner = plannerS2_2
    print("Problem </pro>" + parser.problem_name + "</> solved by System </sys>" + str(systemTWO) + "</> using planner </pla>" + str(planner) +"</>.")
    sys.exit(0)


    if planner == plannerS2_1:
        domainNamePDKB, problemNamePDKB = parser.print_PDKB(output_folderPDKB)
        result = subprocess.run(['sh','./'+ scripts_folder + 'PDKB_solve.sh', problemNamePDKB,  " " + output_folderPDKB, " " + str(int(timeLimit))+"s"])
        solutionS2 = readSolutionFromFile("tmp/PDKB/" + problemNamePDKB)
        time = readTimeFromFile("tmp/PDKB/" + problemNamePDKB)
    elif planner == plannerS2_2:
        result = subprocess.run(['sh','./'+ scripts_folder + 'EFP_solve.sh', instanceNameEFP, " " + output_folderEFP, " " + str(int(timeLimit))+"s"])
        solutionS2 = readSolutionFromFile("tmp/EFP/" + instanceNameEFP)
        time = readTimeFromFile("tmp/EFP/" + instanceNameEFP)
    else:
        raise Exception('Planner '+ str(planner) +' is not a known S2-planner')

    if(time == "TO" or solutionS2 == "noSolution"):
        #return False, timeLimit, None
        print("Problem </pro>" + parser.problem_name + "</> could not be solved by System </sys>" + str(systemTWO) + "</> using planner </pla>" + str(planner) +"</>.")
        return False
    else:
        memorizeSolution(systemTWO, planner, 1.0, float(time), 1.0, solutionS2)
        #return True, float(time), solutionS2

def memorizeSolution(system, planner, confidence, elapsedTime, correctness, solution):


    json_path = dbFolder+jsonFilename

    if not os.path.exists(json_path):
        open(json_path, 'w').close()

    memory_file = open(json_path)

    data = json.load(memory_file)

    if ('cases' not in data.keys()) or (len(data['cases']) == 0):
        index = 0
    else:
        index = max( [int(k) for k in data['cases'].keys()] )

    index += 1

    #jsonformat established here

    data['cases'][str(index)] = {}
    data['cases'][str(index)]['domain_name'] = parser.domain_name
    data['cases'][str(index)]['problem_name'] = parser.problem_name
    data['cases'][str(index)]['difficulty'] = difficulty
    data['cases'][str(index)]['system'] = system
    data['cases'][str(index)]['planner'] = planner
    data['cases'][str(index)]['confidence'] = confidence
    data['cases'][str(index)]['correctness'] = correctness
    data['cases'][str(index)]['solving_time'] = elapsedTime
    data['cases'][str(index)]['total_time'] = time.time()-timeSTART

    init,goal = getStates.States(problemFile) #reading initial and goal states from problem file

    data['cases'][str(index)]['init'] = init
    data['cases'][str(index)]['goal'] = goal
    data['cases'][str(index)]['plan'] = solution

    json_object = json.dumps(data,indent=4)

    # writing dictionary data into json file
    with open(json_path,"w") as out:
        out.write(json_object)


    print("The solution of </pro>" + parser.problem_name + "</> is </sol>" + str(solution) + "</> with correctness </cor>" + str(correctness) + "</> and has been found in </tim>" + str(elapsedTime) + "s</> by System </sys>" + str(system) + "</>", end = '')
    if (system != systemONE):
        print(" using planner </pla>" + str(planner) +"</>.")
    else:
        print(" using mode </mod>" + str(plannerS1) +"</>.")
    sys.exit(0)

def generateNamePDKB():
    domainNamePDKB = EPPDL.getDomainPDKB(domainFile)
    problemNamePDKB = EPPDL.getProblemPDKB(problemFile)

#----NEW

def countSolvedInstances(system,planner):
    dbFilename = dbFolder + jsonFilename
    matchCount = 0
    totalCorrectness = 0


    # Opening JSON file
    try:
        f = open(dbFilename)
        # returns JSON object as a dictionary
        data = json.load(f)

        index = len(data['cases'])
        count = 0
        while (index > 0) and (count < maxExperience):
            if(data['cases'][str(index)]['domain_name'] == parser.domain_name):
                #planner == systemALL means that we accept all the planners
                if(int(data['cases'][str(index)]['system']) == system or system == systemALL):
                    if(int(data['cases'][str(index)]['planner']) == planner or planner == plannerALL):
                        matchCount += 1
            index -= 1
            count += 1

        return matchCount
    except IOError:
        return 0

def getAvgCorrS1(system,planner,slidingWindow):
    dbFilename = dbFolder + jsonFilename
    matchCount = 0
    totalCorrectness = 0

    # Opening JSON file
    f = open(dbFilename)
    # returns JSON object as a dictionary
    data = json.load(f)

    index = len(data['cases'])
    count = 0
    while (index > 0) and (count < maxExperience):
        if(data['cases'][str(index)]['domain_name'] == parser.domain_name):
            #planner == systemALL means that we accept all the planners
            if(int(data['cases'][str(index)]['system']) == system or system == systemALL):
                if(int(data['cases'][str(index)]['planner']) == planner or planner == plannerALL):
                    if (matchCount < slidingWindow):
                        matchCount += 1
                        totalCorrectness += float(data['cases'][str(index)]['correctness'])
                    else:
                        index = 0
        index -= 1
        count += 1


    if matchCount > 0:
        return (totalCorrectness/matchCount)
    else:
        return 0

def tryS1(plannerS1, solutionS1, confidenceS1, timeS1):
    correctnessS1 = validateSolution(solutionS1)
    if (correctnessS1 >= correctnessCntx):
        memorizeSolution(systemONE, plannerS1, confidenceS1, timeS1, correctnessS1, solutionS1)
    else:
        solveWithS2NoPlan(timeLimitCntx)

def selectPlannerS2():
    planner = plannerS2_1 #By default we use plannerS2_1 -- we use label to indicate the planners
    instanceDepth = int(getVarFromFile(problemFile,"depth"))

    if (instanceDepth > maxGroundDepth or checkDCK()): #Check if depth > 3 or if DCK is not required
        planner = plannerS2_2

    return planner
#-----------------------------------------------
# Main
#-----------------------------------------------
if __name__ == '__main__':

    forceP1 = False
    forceP2 = False
    forceDepth = False
    forcedDepth = 0

    for arg in sys.argv:
        if forceDepth and forcedDepth == 0:
            forcedDepth = int(arg)

        if (arg == "-d"):
            forceDepth = True

        if (arg == "P1"):
            forceP1 = True
        elif (arg == "P2"):
            forceP2 = True

    if (len(sys.argv) > 4 and not forceP1 and not forceP2):
        readThreshold(sys.argv[4])



    timeSTART = time.time()
    createFolders()

    domainFile = sys.argv[1]
    problemFile = sys.argv[2]
    context = sys.argv[3]
    #Always used to verify the solution anyway
    generateEFPInstance()

    #correctnessCntx = float(getVarFromFile(context,"correctness"))
    reduced_risk_adversion = 0.4
    correctnessCntx = threshold3 - reduced_risk_adversion
    timeLimitCntx = float(getVarFromFile(context,"timelimit"))
    instanceDepth = int(getVarFromFile(problemFile,"depth"))
    if forcedDepth > 0:
        instanceDepth = forcedDepth

    #print("Instance Depth is " + str(instanceDepth))

    if (forceP1):
        solveWithS2(timeLimitCntx - (time.time() - timeSTART),plannerS2_1)
        sys.exit()
    elif (forceP2):
        solveWithS2(timeLimitCntx - (time.time() - timeSTART),plannerS2_2)
        sys.exit()



    difficulty = estimateDifficulty(instanceDepth)



    ######### S1 metacognitive part
    # AUTOMATICALLY CALL S1
    plannerS1 = plannerS1_Plansformer
    timeS1 = time.time()
    solutionS1, confidenceS1 = solveWithS1()
    timeS1 = time.time() - timeS1
    M = 1



    # IF NUMBER OF CASES IN JSON FILE (that match the domain) IS > THRESHOLD1 THEN CONTINUE WITH TO S1
    if (countSolvedInstances(systemALL,plannerALL) > threshold1):
        if (countSolvedInstances(systemONE,plannerALL) > threshold4):
            M = 1-getAvgCorrS1(systemONE,plannerALL,threshold4)
        else:
            M = 0

        M = 0

        if (confidenceS1 * (1-M) > threshold3):
            tryS1(plannerS1, solutionS1, confidenceS1, timeS1)


    # ######### S2 metacognitive part
    # # Employ the S2 metacognitive structure



    plannerS2 = selectPlannerS2()
    estimatedTimeS2 = estimateTimeCons(plannerS2, difficulty)
    estimatedCostS2 = sys.maxsize

    remainingTime = timeLimitCntx - (time.time() - timeSTART)
    if(remainingTime - estimatedTimeS2 > 0):
        estimatedCostS2 = estimatedTimeS2 / remainingTime

    if (estimatedCostS2 > 1):
        tryS1(plannerS1, solutionS1, confidenceS1, timeS1)
    else:
        probabilityS1 = (1-threshold3)*epsilonS1
        #random() genrates a number between 0 and 1
        if (probabilityS1 > random.random()):
            tryS1(plannerS1, solutionS1, confidenceS1, timeS1)
        else:
            correctnessS1 = validateSolution(solutionS1)
            if (correctnessS1 >= correctnessCntx):
                if((1-(estimatedCostS2 * (1-threshold3))) > (correctnessS1*(1-M))):
                    if (not solveWithS2(remainingTime,plannerS2)):
                        memorizeSolution(systemONE, plannerS1, confidenceS1, timeS1, correctnessS1, solutionS1)
                else:
                    memorizeSolution(systemONE, plannerS1, confidenceS1, timeS1, correctnessS1, solutionS1)
            else:
                solveWithS2(remainingTime,plannerS2)

    print("Problem </pro>" + parser.problem_name + "</> could not be solved by System </sys>Metacognitive</>.")



"""Module to compile an multi agent epistemic planning problem into epddl"""
from typing import Tuple, Union, List
import sys
import os
import string
import random
CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, os.path.join(CURRENT_DIR, '..', '..'))

from model import *

def get_random_string(length: int) -> str:
    """choose from all lowercase letter"""
    letters = string.ascii_lowercase
    result_str = ''.join(random.choice(letters) for i in range(length))
    return result_str


def predicate_argument(fluent_type: Union[IntType,
                                          EnumType,
                                          StructType]) -> str:
    """from FType to epddl argument of predicates"""
    if fluent_type.is_agent_type():
        return f'?ag_{get_random_string(5)} - ' + 'agent'
    type_name = fluent_type.name
    if fluent_type.is_int_type() or fluent_type.is_enum_type():
        return f'?{type_name}_{get_random_string(5)} - ' + type_name
    if fluent_type.is_struct_type():
        return ' '.join(map(predicate_argument, fluent_type))
    return ''


def predicate(fluent: Fluent) -> str:
    """from Fluent to epddl predicate"""
    if fluent.type.is_bool_type():
        return f'{fluent.name}'
    return f'{fluent.name} ' + ' ' + predicate_argument(fluent.type)


def var_param(variable: Variable) -> str:
    """replace agents with agent"""
    if variable.is_agent():
        return f'?{variable.name.lower()} - agent'
    return f'?{variable.name.lower()} - {variable.type.name.lower()}'


def parameters(variables: List[Variable]) -> str:
    """from list of Variable to epddl parameters"""
    variables = variables
    return ' '.join([var_param(variable)
                     for variable in variables])


def param_val_to_epddl(param_val: Union[ArithmeticExpr,
                                        int, Variable, str],
                       prob) -> str:
    """from ArithmeticExpr, int, Variable, str to epddl parameters"""
    if isinstance(param_val, str):
        return param_val
    if isinstance(param_val, int):
        return str(param_val)
    if isinstance(param_val, Variable):
        if prob:
            return f'{param_val.name.lower()}'
        return f'?{param_val.name.lower()}'
    if isinstance(param_val, ArithmeticExpr):
        left_val = param_val_to_epddl(param_val.values[0], prob)
        right_val = param_val_to_epddl(param_val.values[1], prob)
        op = param_val.operator.value
        return '(' + left_val + op + right_val + ')'


def literal(lit: Literal, prob: bool) -> str:
    """from Literal to epddl literal"""
    f_name = f"{lit.fluent.name}"
    return f_name + ' ' + ' '.join(map(lambda x: param_val_to_epddl(x, prob)
                                       , lit.args))


def agent(ag: Union[str, Variable], prob=False) -> str:
    """from agento to EPDDL agent"""
    if isinstance(ag, str):
        if prob:
            return ag
        return '?'+ag
    if isinstance(ag, Variable):
        if prob:
            return ag.name.lower()
        return '?'+ag.name.lower()


def negate(p: str, prob: bool) -> str:
    """neagte a predicate.If prob for the domain enc., else for problem enc."""
    if prob:
        return 'not (' + p + ')'
    return '-' + p


def pred(_predicate: Predicate, prob=False) -> str:
    """from Predicate to epddl's action predicate"""
    neg = _predicate.negated
    epddl_pred = ''

    if isinstance(_predicate, Literal):
        epddl_pred = f'{literal(_predicate, prob)}'

    elif isinstance(_predicate, BeliefPredicate):
        def to_agent(ag: Union[str, Variable]):
            return agent(ag, prob)
        
        agents = '[' + ' '.join(map(to_agent, _predicate.agents)) + ']'
        prop = pred(_predicate.belief_proposition, prob)
        epddl_pred = f'{agents}({prop})'

    elif isinstance(_predicate, BooleanPredicate):
        op = _predicate.op.value
        l_pred = pred(_predicate.left_predicate, prob)
        r_pred = pred(_predicate.right_predicate, prob)
        epddl_pred = op + ' (' + l_pred + ') (' + r_pred + ')'

    elif isinstance(_predicate, When):
        return f'when ({pred(_predicate.body, prob)}) ({pred(_predicate.head, prob)})'

    elif isinstance(_predicate, EqualityPredicate):
        op = _predicate.operator.value
        l_pred = predicate_argument(_predicate.left_operand, prob)
        r_pred = predicate_argument(_predicate.right_operand, prob)
        return op + ' (' + l_pred + ') (' + r_pred + ')'

    if neg: epddl_pred = negate(epddl_pred, prob=prob)

    return epddl_pred


def preds(predicates: List[Predicate]) -> str:
    """from list of Predicate to epddl's action predicates"""
    if len(predicates) == 1:
        return f'{pred(predicates[0])}'
    return 'and ' + ' '.join(map(lambda x: '(' + x + ')',
                                 map(pred, predicates)))


def obss(observers: List[Union[Forall, str, Variable]]) -> str:
    """from list of observer to epddl observers"""
    return ' '.join(map(obs, observers))


def obs(observer: Union[Forall, str, Variable]) -> str:

    print(type(observer))
    print(isinstance(observer, Variable))

    if isinstance(observer, str):
        return f'({agent(observer)})'

    if isinstance(observer, Variable):
        print(f'({agent(observer.name.lower())})')
        return f'({agent(observer.name.lower())})'

    if isinstance(observer, Forall):
        who = f'{agent(observer.who)}'
        if observer.when is not None:
            who = f'when ({pred(observer.when)}) ({who})'

        forall = who
        if observer.disequality_predicate is not None:
            neq = observer.disequality_predicate
            agentl = agent(neq.left_operand.name.lower())
            agentr = agent(neq.right_operand.name.lower())
            if observer.who == neq.left_operand:
                forall = f'diff ({agentl}) ({agentr})'
            else:
                forall = f'diff ({agentr}) ({agentl})'

        return f'(forall ({forall}) ({who}))'


def action(mep_action: MEAction) -> str:
    """from MEAction to epddl action"""
    action_enc = f'\t(:action {mep_action.name}\n'
    action_enc += f'\t\t:act_type {mep_action.type.value}\n'
    if len(mep_action.params) > 0:
        action_enc += f'\t\t:parameters ({parameters(mep_action.params)})\n'
    if len(mep_action.precondition) > 0:
        action_enc += f'\t\t:precondition ({preds(mep_action.precondition)})\n'
    if len(mep_action.effects) > 0:
        action_enc += f'\t\t:effect ({preds(mep_action.effects)})\n'
    if len(mep_action.full_obs) > 0:
        print(mep_action.full_obs)
        action_enc += f'\t\t:observers (and {obss(mep_action.full_obs)})\n'
    if len(mep_action.partial_obs) > 0:
        action_enc += f'\t\t:p_observers (and {obss(mep_action.partial_obs)})'
    return action_enc + '\n\t)'


def agent_names(problem: MEPlanningProblem) -> str:
    """serach for the type of agents and returns a list of agent names"""
    agent_type = next(iter([t for t in problem.types
                            if t.is_agent_type()]))
    return ' '.join(agent_type.domain)


def enclose_into_brackets(p: str) -> str:
    return '(' + p + ')'


def problem_predicate_encoding(p: Predicate) -> str:
    return pred(p, prob=True)


def init(problem: MEPlanningProblem) -> str:
    """obtain sequence of epddl initial predicates"""
    return ' '.join(map(enclose_into_brackets,
                        map(problem_predicate_encoding,
                            problem.init_values)))

def goal(problem: MEPlanningProblem) -> str:
    """obtain sequence of epddl goal predicates"""
    return ' '.join(map(enclose_into_brackets,
                        map(problem_predicate_encoding,
                            problem.goals)))

#  TODO till now only enums are supported
def process_type(t: Type) -> str:
    return '\t\t' + ' '.join(t.domain) + ' - ' + t.name.lower()


def types(problem: MEPlanningProblem) -> str:
    """from types to PDDL types"""
    if len(problem.types) > 1:
        types = map(process_type, [t for t in problem.types if not t.is_agent_type() and not t.is_struct_type()])
        return '\t(:objects\n' + '\n'.join(types) + '\n\t)\n'
    return ''


def compile_into_epddl(problem: MEPlanningProblem) -> Tuple[str, str]:
    """from pd.MEPlanningProblem to epddl domain and problem files"""
    p_dom = f'(define (domain {problem.name})\n'
    p_dom += '\t(:requirements :strips :typing :negative-preconditions :mep :no-duplicates)\n'
    #  fluents:
    p_dom += '\t(:predicates '
    p_dom += ' '.join(map(lambda fluent: f'({predicate(fluent)})',
                          problem.fluents)) + ')\n'

    # actions:
    p_dom += '\n'.join(map(action, problem.actions))

    p_dom += '\n)\n'
    p_inst = f'(define (problem {problem.name})\n'
    p_inst += types(problem)
    p_inst += f'\t(:domain {problem.name})\n'
    p_inst += f'\t(:agents {agent_names(problem)})\n'
    p_inst += '\t(:depth 2)\n'
    p_inst += '\t(:dynck false)\n'
    p_inst += f'\t(:init {init(problem)})\n'
    p_inst += f'\t(:goal {goal(problem)})\n'
    p_inst += ')'
    return p_dom, p_inst



"""This modul contains Action datatypes"""
import os, sys
from typing import List, Union, TypeVar
from dataclasses import dataclass
from enum import Enum
from collections import namedtuple

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, current_dir)

from .fluent import Fluent
from .predicate import Literal, Predicate, BeliefPredicate, Forall
from .variable import Variable
from .goal import Poset

#  TODO: introduction of templates
#      needs of a fluent definition wich can be or
#      boolean or of a particular type
#      you should also be able to define new types
#  TODO: definition of a fixed set of actions that
#      have a direct encoding into moveit



@dataclass(frozen=True)
class IAction():
    """IAction class introduced to represent classical planning actions"""
    name: str
    params: List[Variable]
    precondition: List[Predicate]
    effects: List[Literal]

    def __post_init__(self):

        for param in self.params:
            assert isinstance(param, Variable)
        for precond in self.precondition:
            assert isinstance(precond, Predicate)
        for effect in self.effects:
            assert isinstance(effect, Literal)

    def __repr__(self):
        return self.name + '(' + str(self.params) + ')'


@dataclass(frozen=True)
class MEActionType(Enum):
    """Type of Multi-Agent Epistemic Actions"""
    ontic = 'ontic'
    sensing = 'sensing'
    announcement = 'announcement'


@dataclass#(frozen=True)
class MEAction():
    """MEAction class introduced to represent Mulit-Agent Epistemic Actions"""
    
    name: str
    params: List[Variable]
    type: MEActionType
    precondition: List[Predicate]
    effects: List[Predicate]
    full_obs: List[Union[str, Variable, Forall]]
    partial_obs: List[Union[str, Variable, Forall]]

    def __init__(self, name, params = [], type = MEActionType.ontic,
                 precondition = [], effects = [], full_obs = [], partial_obs = []):
        self.name = name
        self.params = params
        self.type = type
        self.precondition = precondition
        self.effects = effects
        self.full_obs = full_obs
        self.partial_obs = partial_obs
    
    def __post_init__(self):

        for param in self.params:
            assert isinstance(param, Variable)
        for precond in self.precondition:
            assert isinstance(precond, Predicate)
        for effect in self.effects:
            assert isinstance(effect, Predicate)
        for obs in self.full_obs + self.partial_obs:
            assert isinstance(obs, (str, Variable, Forall))

    def classical_sub_goals(self, sub_goal: Union[Poset, List[Literal]]):
        #only ontic actions can be refined:
        assert self.type == MEActionType.ontic
        self._sub_goal = sub_goal

    @property
    def sub_goals(self) -> Union[Poset, List[Literal]]:
        return self._sub_goal

Instantiated_Action = namedtuple('Instantiated_Action',
                                 ['action', 'var_map'])



"""
module for definition of Variable and ArithmeticExpr
which are the possible arguments of a predicate
"""
import sys, os
from dataclasses import dataclass

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, current_dir)

from .ftype import Type, BoolType, AgentType
from .arithmetic_expression import ArithmeticExpr, is_int_fvalue, \
    ArithmeticOperator


@dataclass
class Variable():
    """Variable of a given ftype.Type"""
    def __init__(self, name: str, vtype: Type, agent=False):
        if isinstance(vtype, BoolType):
            raise Exception("""do not define a varible boolean,
            define instead a boolean fluent""")
        self._name = name
        self._vtype = vtype
        self._agent = agent

    def is_agent(self) -> bool:
        """Returns True iff the variable represents an agent"""
        return isinstance(self._vtype, AgentType)

    @property
    def name(self):
        """Get the name of the variable"""
        return self._name

    @property
    def type(self):
        """Get the ftype.Typeof the variable"""
        return self._vtype

    def __hash__(self):
        return hash(self.name)#**hash(self.type)

    def __eq__(self, var):
        return var.name == self.name and var.type == self.type

    def __repr__(self):
        return f"var {self.name} of type {self.type}"

    def __str__(self):
        return self.name

    def _op_int(self, other, operator):
        """Return an arithmetic expression"""
        if self.type.is_int_type():
            if is_int_fvalue(other):
                return ArithmeticExpr(op=operator, values=[self, other])
        raise Exception("only integer varialbes can be used as sum term")

    def __add__(self, other):
        return self._op_int(other, ArithmeticOperator.Plus)

    def __radd__(self, other):
        return self._op_int(other, ArithmeticOperator.Plus)

    def __sub__(self, other):
        return self._op_int(other, ArithmeticOperator.Sub)

    def __rsub__(self, other):
        return self._op_int(other, ArithmeticOperator.Sub)



"""This module contain the definition of Fluent"""
import sys, os
from dataclasses import dataclass
from typing import ClassVar

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, current_dir)

from .predicate import Literal
from .arithmetic_expression import ArithmeticExpr
from .ftype import BoolType, Type
from .variable import Variable


@dataclass(frozen=True)
class Fluent():
    """Fluent class to represnet classical planning fluents"""
    default_type : ClassVar = BoolType('bool')
    name: str
    type: Type = default_type

    def __post_init__(self):

        if not isinstance(self.name, str):
            raise Exception("The fluent name must be a string")

        if not isinstance(self.type, Type):
            raise Exception("""The fluent type must be \
                                an instance of fType.Type""")

    def __eq__(self, other):
        return self.name == other.name and self.type == other.type

    def __repr__(self):
        return f"fluent {self.name} of type {self.type}"

    @staticmethod
    def _check_type(arg, _type: Type, pos: int) -> bool:
        """check if the value {arg} has type types[{pos}]"""
        str_error = f"""Fluent has type {_type} at position {pos},\
        but an argument: \"{arg}\" was given"""
        if _type.is_int_type():

            if isinstance(arg, int) and _type.min <= arg <= _type.max:
                return True
            if isinstance(arg, ArithmeticExpr):
                #I do not add other checks
                return True
            if arg.type.contained_in(_type):
                #if it is a variable, its domain should be contained
                return True
    
        if _type.is_enum_type():
            if isinstance(arg, str) and arg in _type.enum_values or \
               arg.type.contained_in(_type):
                return True
            
        if _type.is_bool_type():
            if isinstance(arg, bool) or arg.type.is_bool_type():
                return True

        raise Exception(str_error)

    def __call__(self, *args) -> Literal:

        if len(self.type) != len(args):
            raise Exception(f"""Fluent {self.name} has {len(self.type)} arguments,
                            but {len(args)} were given""")

        vars = []
        for arg in args:
            if isinstance(arg, Variable):
                vars.append(arg)

        if len(self.type) == 0:
            return Literal(negated = False, types = [self.type],
                           variables = [], fluent = self, args = [])

        if len(self.type) == 1:
            self._check_type(args[0], self.type, 0)
            return Literal(negated = False, types = [self.type],
                           variables = vars, fluent = self, args = list(args))

        c = 0
        for arg, t in zip(args, self.type):
            self._check_type(arg, t, c)
            c += 1
        return Literal(negated = False, types = [self.type], variables = vars,
                       fluent = self, args = list(args))



from dataclasses import dataclass

@dataclass(frozen = True)
class Interval:

    min : int
    max : int

    def __init__(self, _min, _max):
        if (_min > _max):
            raise Exception(f"min value {_min} > max value{_max}.\n")
        object.__setattr__(self, "min", _min)
        object.__setattr__(self, "max", _max)

    def overlaps(self, i : 'Interval'):
        return min(self.max, i.max) - max(self.min, i.min) >= 0

    def __eq__(self, i : 'Interval'):
        return self.min == i.min and self.max == i.max
        

def main():
    i = Interval(2, 5)
    j = Interval(3, 6)
    print(i.overlaps(j))
    k = Interval(10, 15)
    print(i.overlaps(k))
    l = Interval(5, 7)
    print(l.overlaps(i))
    print("isinstance")
    print(isinstance(l, Interval))
    Interval(3, 1)
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Four spaces as indentation [no tabs]

import re
import itertools
import warnings
import copy
import os


from pathlib import Path

from .action import Action

DEBUG = False

class EPDDL_Parser:

    SUPPORTED_REQUIREMENTS = [':strips', ':negative-preconditions', ':typing', ':no-duplicates', ':mep']

    #-----------------------------------------------
    # Tokens
    #-----------------------------------------------

    def scan_tokens(self, filename):
        try:
            with open(filename,'r') as f:
                # Remove single line comments
                str = re.sub(r';.*$', '', f.read(), flags=re.MULTILINE).lower()
                str = re.sub(r'\[([^[]+)-agent(\s+|)\]', r'[\1]',str,flags=re.MULTILINE)

                nb_rep = 1

                while (nb_rep):
                        (str, nb_rep) = re.subn(r'\((\s|)+\(([^()]+)\)(\s|)+\)', r'\2',str,flags=re.MULTILINE)

                nb_rep = 1

                while (nb_rep):
                    (str, nb_rep) = re.subn(r'(\[[^[]+\])\(([^(]+)\)', r'\1\2',str,flags=re.MULTILINE)

            # Tokenize
            stack = []
            list = []
            isBF = 0
            insideBF = 0
            firstAg = 1
            countSqPa = 0
            multi_ag = 0
            Bf_string = ''
            for t in re.findall(r'[()\[\]]|[^\s()\[\]]+', str):
                if t == '(':
                    stack.append(list)
                    list = []
                elif t == ')':
                    if stack:
                        l = list
                        list = stack.pop()
                        list.append(l)
                    else:
                        raise Exception('Missing open parentheses')
                elif t == '[':
                    firstAg = 1
                    insideBF = 1
                    Bf_string = 'B('
                elif t == ']':
                    insideBF = 0
                    Bf_string += ','
                    if multi_ag == 1:
                        Bf_string = Bf_string.replace('B(', 'C(')
                    list.append(Bf_string)
                    multi_ag = 0
                elif insideBF == 1:
                    if firstAg == 0:
                        multi_ag = 1
                        Bf_string +=','
                    Bf_string +=t
                    firstAg = 0
                else:
                    list.append(t)
            if stack:
                raise Exception('Missing close parentheses')
            if len(list) != 1:
                raise Exception('Malformed expression')
            return list[0]

        except Exception as e: print(e)

    #-----------------------------------------------
    # Parse domain
    #-----------------------------------------------

    def parse_domain(self, domain_filename):
        tokens = self.scan_tokens(domain_filename)
        if type(tokens) is list and tokens.pop(0) == 'define':
            self.domain_name = 'unknown'
            self.requirements = []
            self.types = {}
            self.objects = {}
            self.actions = []
            self.predicates = {}
            while tokens:
                group = tokens.pop(0)
                t = group.pop(0)
                if t == 'domain':
                    self.domain_name = group[0]
                elif t == ':requirements':
                    for req in group:
                        if not req in self.SUPPORTED_REQUIREMENTS:
                            raise Exception('Requirement ' + req + ' not supported')
                    self.requirements = group
                elif t == ':constants':
                    self.parse_objects(group, t)
                elif t == ':predicates':
                    self.parse_predicates(group)
                elif t == ':types':
                    self.parse_types(group)
                elif t == ':action':
                    self.parse_action(group)
                else: self.parse_domain_extended(t, group)
        else:
            raise Exception('File ' + domain_filename + ' does not match domain pattern with \''+str(tokens)+'\'')

    def parse_domain_extended(self, t, group):
        print(str(t) + ' is not recognized in domain')

    #-----------------------------------------------
    # Parse hierarchy
    #-----------------------------------------------

    def parse_hierarchy(self, group, structure, name, redefine):
        list = []
        while group:
            if redefine and group[0] in structure:
                raise Exception('Redefined supertype of ' + group[0])
            elif group[0] == '-':
                if not list:
                    raise Exception('Unexpected hyphen in ' + name)
                group.pop(0)
                type = group.pop(0)
                if not type in structure:
                    structure[type] = []
                structure[type] += list
                list = []
            else:
                list.append(group.pop(0))
        if list:
            if not 'object' in structure:
                structure['object'] = []
            structure['object'] += list

    def parse_hierarchy_ag(self, group, structure, name, redefine):
        list = []
        while group:
            if redefine and group[0] in structure:
                raise Exception('Redefined supertype of ' + group[0])
            elif group[0] == '-':
                raise Exception('Unexpected hyphen in ' + name)
            else:
                list.append(group.pop(0))
        if list:
            if not 'agent' in structure:
                structure['agent'] = []
            structure['agent'] += list

    #-----------------------------------------------
    # Parse objects
    #-----------------------------------------------

    def parse_objects(self, group, name):
        self.parse_hierarchy(group, self.objects, name, False)

    def parse_agents(self, group, name):
        self.parse_hierarchy_ag(group, self.objects, name, False)

    # -----------------------------------------------
    # Parse types
    # -----------------------------------------------

    def parse_types(self, group):
        self.parse_hierarchy(group, self.types, 'types', True)

    #-----------------------------------------------
    # Parse predicates
    #-----------------------------------------------

    def parse_predicates(self, group):
        for pred in group:
            predicate_name = pred.pop(0)
            if predicate_name in self.predicates:
                raise Exception('Predicate ' + predicate_name + ' redefined')
            arguments = {}
            untyped_variables = []
            while pred:
                t = pred.pop(0)
                if t == '-':
                    if not untyped_variables:
                        raise Exception('Unexpected hyphen in predicates')
                    type = pred.pop(0)
                    while untyped_variables:
                        arguments[untyped_variables.pop(0)] = type
                else:
                    untyped_variables.append(t)
            while untyped_variables:
                arguments[untyped_variables.pop(0)] = 'object'
            self.predicates[predicate_name] = arguments



    #-----------------------------------------------
    # Parse action
    #-----------------------------------------------

    def parse_action(self, group):
        name = group.pop(0)
        if not type(name) is str:
            raise Exception('Action without name definition')
        for act in self.actions:
            if act.name == name:
                raise Exception('Action ' + name + ' redefined')
        parameters = []
        act_type = 'ontic'
        positive_preconditions = []
        negative_preconditions = []
        add_effects = []
        del_effects = []
        f_obs = []
        p_obs = []
        derive_cond = []
        explicit_eff = []
        extensions = None
        while group:
            t = group.pop(0)
            if t == ':parameters':
                if not type(group) is list:
                    raise Exception('Error with ' + name + ' parameters')
                parameters = []
                untyped_parameters = []
                p = group.pop(0)
                while p:
                    t = p.pop(0)
                    if t == '-':
                        if not untyped_parameters:
                            raise Exception('Unexpected hyphen in ' + name + ' parameters')
                        ptype = p.pop(0)
                        while untyped_parameters:
                            parameters.append([untyped_parameters.pop(0), ptype])
                    else:
                        untyped_parameters.append(t)
                while untyped_parameters:
                    parameters.append([untyped_parameters.pop(0), 'object'])
            elif t == ':act_type':
                act_type = self.assign_act_type(group.pop(0))
            elif t == ':precondition':
                self.split_predicates(group.pop(0), positive_preconditions, negative_preconditions, name, ' preconditions')
            elif t == ':effect':
                #self.split_effects(group.pop(0), add_effects, del_effects, name, ' effects')
                self.recoursive_reading(group.pop(0), [['']], [['']], [['']], 0, add_effects, del_effects, name, ' effects')

            #    print(str([list(i) for i in add_effects]))
            #    print(str([list(i) for i in del_effects]))
            elif t == ':observers':
                #self.read_observer(group.pop(0), f_obs, name, ' agents')
                self.recoursive_reading(group.pop(0), [['']], [['']], [['']], 0, f_obs, [], name, ' agents')

            elif t == ':p_observers':
                self.recoursive_reading(group.pop(0), [['']], [['']], [['']], 0, p_obs, [], name, ' agents')
            elif t == ":derive":
                derive_cond = group.pop(0)
            elif t == ":exp_effect":
                explicit_eff = group.pop(0)
            else: extensions = self.parse_action_extended(t, group)
        self.actions.append(Action(name, act_type, parameters, positive_preconditions, negative_preconditions, add_effects, del_effects, f_obs, p_obs, derive_cond, explicit_eff, extensions))

    def parse_action_extended(self, t, group):
        print(str(t) + ' is not recognized in action')

    #-----------------------------------------------
    # Parse problem
    #-----------------------------------------------

    def parse_problem(self, problem_filename):

        #Default depth value
        self.depth = 2

        def frozenset_of_tuples(data):
            return frozenset([tuple(t) for t in data])
        tokens = self.scan_tokens(problem_filename)
        if type(tokens) is list and tokens.pop(0) == 'define':
            self.problem_name = 'unknown'
            self.state = frozenset()
            self.positive_goals = frozenset()
            self.negative_goals = frozenset()
            self.dynamicCK = "false"
            while tokens:
                group = tokens.pop(0)
                t = group.pop(0)
                if t == 'problem':
                    self.problem_name = group[0]
                elif t == ':domain':
                    if self.domain_name != group[0]:
                        raise Exception('Different domain specified in problem file')
                elif t == ':requirements':
                    pass # Ignore requirements in problem, parse them in the domain
                elif t == ':objects':
                    self.parse_objects(group, t)
                elif t == ':agents':
                    self.parse_agents(group, t)
                elif t == ':depth':
                    self.depth = group[0]
                elif t == ':dynck':
                    self.dynamicCK = group[0]
                elif t == ':init':
                    init = []
                #    tmp_group = []
                #    tmp_group.insert(0, 'and')
                #    tmp_group.insert(1, group)
                    group.insert(0,'and')
                    self.split_predicates(group, init, [], '', 'init')
                    self.state = init
                elif t == ':goal':
                    positive_goals = []
                    negative_goals = []
                    group.insert(0,'and')
                    self.split_predicates(group, positive_goals, negative_goals, '', 'goals')
                    self.positive_goals = positive_goals
                    self.negative_goals = negative_goals
                else: self.parse_problem_extended(t, group)
        else:
            raise Exception('File ' + problem_filename + ' does not match problem pattern')

    def parse_problem_extended(self, t, group):
        print(str(t) + ' is not recognized in problem')

    #-----------------------------------------------
    # Split predicates
    #-----------------------------------------------

    def split_predicates(self, group, positive, negative, name, part):
        if not type(group) is list:
            raise Exception('Error with ' + name + part)
        if group[0] == 'and':
            group.pop(0)
        else:
            group = [group]
        for predicate in group:
            #print("Here with predicate[0] == " + str(predicate[0]))
            if 'B(' in predicate[0] or 'C(' in predicate[0]:
                if type(predicate[1]) is list:
                    if predicate[1][0] == 'not':
                        if len(predicate[1][1]) > 0:
                            i = 0
                            tmp_predicate=[]
                            tmp_predicate.insert(0,predicate[0])
                            while i < len(predicate[1][1]):
                                if (i == 0):
                                    tmp_predicate.insert(i+1,'-'+predicate[1][1][0])
                                else:
                                    tmp_predicate.insert(i+1,predicate[1][1][i])
                                i = i+1
                            predicate = tmp_predicate
                        else:
                            raise Exception('Expected predicate after a \'not\'')

            if predicate[0] == 'not':
                if len(predicate) != 2:
                    raise Exception('Unexpected not in ' + name + part)
                negative.append(predicate[-1])
                #print("Appended " + str(predicate[-1]))
            else:
                positive.append(predicate)

    def recoursive_reading(self, body, head_positive, head_negative, diff, subProcedure, positive, negative, name, part):
        if not type(body) is list:
            raise Exception('Error with ' + name + part)

        if body[0] == 'and':
            body.pop(0)
            and_count = 0
            total_body = []
            while and_count < len(body):
                total_body.append(self.recoursive_reading(body[and_count], head_positive, head_negative, diff, subProcedure, positive, negative, name, part))
                and_count = and_count + 1

            #print("Total body: " + str(total_body))
            ret = ([],[])
            for elem in total_body:
                if elem:
                #    print("Elem: " + str(elem))
                    if elem[1] == 0:
                        ret[0].append(elem[0])
                    else:
                        ret[1].append(elem[0])

            return ret


        elif body[0] == 'when':
            body.pop(0)
            condition = body[0]
            body.pop(0)
            #if type(condition) is list:
            if (condition[0] == 'when' or condition[0] == 'forall'):
                raise Exception('Error with ' + name + part + ' you cannot embed other keywords, other than \'and\', in the \'when\' condition')
            elif condition[0] == 'and':
                condition = self.recoursive_reading(condition, [['']], [['']], [['']], 1, positive, negative, name, part)
                pos_condition = condition[0]
                neg_condition = condition[1]

            elif condition[0] == 'not':
                condition.pop(0)
                neg_condition = condition
                pos_condition = [['']]

            else:
                pos_condition = [condition]
                neg_condition = [['']]

            rule = body[0]
            body.pop(0)
            if (rule[0] == 'when' or rule[0] == 'forall'):
                raise Exception('Error with ' + name + part + ' you cannot embed other keywords, other than \'and\', in the \'when\' body')

            self.recoursive_reading(rule,pos_condition,neg_condition, diff, subProcedure, positive, negative, name, part)
            return(rule,pos_condition,neg_condition)

        elif body[0] == 'forall':
            if part != ' agents':
                raise Exception('\'Forall\' keyword only implemented for agents')
            else:
                body.pop(0)
                head = body[0]
                body.pop(0)
                #if type(condition) is list:
                #make sense inside forall
                if head[0] == 'diff':
                    head.pop(0)
                    if len(head) != 2:
                        raise Exception('Bad \'diff\' construction')
                    else:
                        diff = [head[1]]
                        head = head[0]

                if (head[0] == 'when' or head[0] == 'forall' or head[0] == 'and' or head[0] == 'not'):
                    raise Exception('Error with ' + name + part + ' you cannot embed other keywords in the \'forall\' condition')
                else:
                    fa_start = "FASTART"
                    fa_stop = "FASTOP"
                    rule = body[0]
                    body.pop(0)
                    for v in head:
                        if '?' in v:
                            if v in rule:
                                rule[rule.index(v)] =  fa_start + rule[rule.index(v)] + fa_stop
                                self.recoursive_reading(rule,[['']], [['']],[['']], subProcedure, positive, negative, name, part)
                            elif rule[0] == 'when':
                                parsed_rule = self.recoursive_reading(rule,[['']], [['']],[['']], 1, positive, negative, name, part)

                                i = 0
                                while i < 3:
                                    if i > 0:
                                        j = 0
                                        while j < len(parsed_rule[i]):
                                            if v in parsed_rule[i][j]:
                                                parsed_rule[i][j][parsed_rule[i][j].index(v)] =  fa_start + parsed_rule[i][j][parsed_rule[i][j].index(v)] + fa_stop
                                            j = j+1
                                    else:
                                        if v in parsed_rule[i]:
                                            parsed_rule[i][parsed_rule[i].index(v)] =  fa_start + parsed_rule[i][parsed_rule[i].index(v)] + fa_stop
                                    i = i+1
                                self.recoursive_reading(parsed_rule[0],parsed_rule[1],parsed_rule[2], diff, subProcedure, positive, negative, name, part)
                            else:
                                raise Exception('To many nested command in the agents\' observability\n' + str(rule))


        elif body[0] == 'not':
            if len(body) != 2:
                raise Exception('Unexpected not in ' + name + part)
            if subProcedure == 0:
                negative.append((body[-1], head_positive, head_negative,diff))
            return (body[-1], 1)

        else:
            if subProcedure == 0:
                positive.append((body, head_positive, head_negative,diff))
            return (body, 0)

    def assign_act_type(self, name):
        name = name.lower()
        if name == 'ontic' or name == 'announcement' or name == 'sensing':
            return name.lower()
        else:
            raise Exception('Error with the action type definition. Please select one of the following: \'ontic\', \'sensing\', \'announcement\'')


    #-----------------------------------------------
    # Print EFP
    #-----------------------------------------------
    def print_EFP(self,output_folder, file_path=None):
        #########File NAME
        Path(output_folder).mkdir(parents=True,exist_ok=True)




        file_name = self.domain_name + '_' + self.problem_name
        if file_path is None:
            out = open(output_folder +  file_name+".tmp", "w")
        else:
            out = open(file_path, 'w')
        out.write("%This file is automatically generated from an E-PDDL specification and follows the mAp syntax.\n\n")

        #Generate grounded actions and add grounded fluents
        fluents = set()
        ground_actions = []
        for action in self.actions:
            for act in action.groundify(self.objects, self.types, self.requirements, fluents):
                act_name = act.name
                for parameter in act.parameters:
                    act_name += '_'+parameter
                act.name = act_name
                ground_actions.append(act)
        #########FLuents
        self.generate_fluents_EFP(fluents)
        if '' in fluents:
            fluents.remove('')
        out.write('%%%%%%%%%%%%%%%%%%%%%%%%%    FLUENTS    %%%%%%%%%%%%%%%%%%%%%%%%\n')
        out.write('%Fluents generated from EPDDL by grounding each predicate (and cheking in :init, :goal and actions for extra predicates)\n')
        out.write('%The fluents are lexicographically sorted and printed in sets of 10\n\n')
        out.write('fluent ')
        fl_count = 0
        for fluent in sorted(fluents):
            out.write(str(fluent))
            if (fl_count != len(fluents)-1):
                if((fl_count+1)%10 == 0):
                    out.write(';\nfluent ')
                else:
                    out.write(', ')
                fl_count +=1
        self.fl_number = (fl_count +1)
        out.write(';\n\n')
        out.write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n\n')

        #########Actions Names
        out.write('%%%%%%%%%%%%%%%%%%%%%    ACTIONS\' NAMES    %%%%%%%%%%%%%%%%%%%%%\n')
        out.write('%Actions\' names generated from EPDDL by adding to each action names its grounded predicates\n\n')
        out.write('action ')
        act_count = 0
        for action in ground_actions:
            out.write(action.name)
            if (act_count != len(ground_actions)-1):
                if((act_count+1)%10 == 0):
                    out.write(';\naction ')
                else:
                    out.write(', ')
                act_count +=1
        self.act_number = (act_count +1)
        out.write(';\n\n')
        out.write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n\n')


        out.write('%%%%%%%%%%%%%%%%%%%%%    AGENTS\' NAMES    %%%%%%%%%%%%%%%%%%%%%%\n')
        out.write('%Agents\' names generated from EPDDL by looking at the \'agent\' predicate\n\n')
        out.write('agent ')
        ag_count = 0
        for agent in self.objects['agent']:
            out.write(agent)
            if (ag_count != len(self.objects['agent'])-1):
                if((ag_count+1)%10 == 0):
                    out.write(';\nagent ')
                else:
                    out.write(', ')
                ag_count +=1
        self.ag_number = (ag_count + 1)
        out.write(';\n\n')
        out.write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n\n')


        #########Actions Specifications
        out.write('%%%%%%%%%%%%%%%%%    ACTIONS\' SPECIFICATIONS    %%%%%%%%%%%%%%%%\n')
        out.write('%Actions\' specifications generated from EPDDL by grounding each action\'s definition\n\n')
        for action in ground_actions:
            out.write('%%%Action ' + action.name + '\n\n')
            out.write('executable ' + action.name)
            self.print_precondition_EFP(action, out)
            self.print_effects_EFP(action, out)
            self.print_observers_EFP(action, 1, out)
            self.print_observers_EFP(action, 0, out)
            out.write('\n%%%\n\n')
        out.write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n\n')

        #########Actions Specifications
        out.write('%%%%%%%%%%%%%%%%%%    INITIAL FLUENTS TRUTH   %%%%%%%%%%%%%%%%%%\n')
        out.write('%Fluents are considered true when are inserted in :init; otherwise are considered false\n\n')
        out.write('%%%True fluents\n')
        out.write('initially ')
        ini_count = 0
        true_fluents = set()
        belief_ini= set()
        temp_ini = list(self.state)
        for index, ini_f in enumerate(temp_ini):
            ini_fs = self.unify_fluent_EFP(ini_f)
            if 'B(' in ini_fs or 'C(' in ini_fs:
                belief_ini.add(ini_fs)
            else:
                out.write(ini_fs)
                true_fluents.add(ini_fs)
                if ( (index+1 < len(temp_ini)) and ('B(' not in temp_ini[index+1][0]  and 'C(' not in temp_ini[index+1][0])):
                    out.write(', ')
        out.write(';\n')
        neg_fluents = fluents - true_fluents

        out.write('%%%False fluents\n')
        out.write('initially ')
        ini_count = 0
        for ini_f in neg_fluents:
            out.write('-'+ini_f)
            if (ini_count != len(neg_fluents)-1):
                out.write(', ')
                ini_count+=1
        out.write(';\n')
        out.write('\n')
        out.write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n\n')


        out.write('%%%%%%%%%%%%%%%%%%    INITIAL BELIEFS TRUTH   %%%%%%%%%%%%%%%%%%\n')
        out.write('%Extracted from the :init field\n\n')
        ini_count = 0
        for ini_bf in belief_ini:
            out.write('initially ')
            out.write(ini_bf)
            if (ini_count != len(belief_ini)-1):
                out.write(';\n')
                ini_count+=1
        if (len(belief_ini) > 0):
            out.write(';\n')
        out.write('\n')
        out.write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n\n')

        out.write('%%%%%%%%%%%%%%%%%%%%%%%%%%    GOALS   %%%%%%%%%%%%%%%%%%%%%%%%%%\n')
        out.write('%The goals of the plan. Each goal is presented separately to ease the reading\n\n')
        for goal_f in self.positive_goals:
            out.write('goal ')
            goal_fs = self.unify_fluent_EFP(goal_f)
            out.write(goal_fs + ';\n')

        for goal_f in self.negative_goals:
            out.write('goal (-')
            goal_fs = self.unify_fluent_EFP(goal_f)
            out.write(goal_fs + ');\n')

        out.write('\n')
    #    out.write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n')
        out.close()
        return file_name+".tmp"

    def unify_fluent_EFP(self,given_list):
        return Action.unify_fluent_EFP(given_list)

    def generate_fluents_EFP(self, fluents_set):

        for ini_f in self.state:
            fluent = self.unify_fluent_EFP(ini_f)
            if 'B(' not in fluent and 'C(' not in fluent and not '-' in fluent:
                fluents_set.add(fluent)

        for goal_f in self.positive_goals:
            fluent = self.unify_fluent_EFP(goal_f)
            if 'B(' not in fluent and 'C(' not in fluent and not '-' in fluent:
                fluents_set.add(fluent)

        for goal_f in self.negative_goals:
            fluent = self.unify_fluent_EFP(goal_f)
            if 'B(' not in fluent and 'C(' not in fluent and not '-' in fluent:
                fluents_set.add(fluent)

        #duplicates = True
        duplicates = False
        if ':no-duplicates' in self.requirements:
            duplicates = False
        for predicate in self.predicates.items():
            #print('original:' + str(predicate))
            type_map = []
            variables = []
            pred_ini=[]
            pred_ini.append(predicate[0])
            for var in self.predicates[predicate[0]]:
                type = self.predicates[predicate[0]][var]
                #print ('Type: ' + str(type) + ' var: ' + var + ' predicate: ' + predicate[0])
                pred_ini.append(var)
                type_stack = [type]
                items = []
                while type_stack:
                    t = type_stack.pop()
                    if t in self.objects:
                        items += self.objects[t]
                    elif t in self.types:
                        type_stack += self.types[t]
                    else:
                        raise Exception('Unrecognized type ' + t)
                type_map.append(items)
                variables.append(var)
            for assignment in itertools.product(*type_map):
                if (not duplicates and len(assignment) == len(set(assignment))) or duplicates:
                    #pred = predicate
                    pred = list(pred_ini)
                    iv = 0
                #    print(str(variables))
                #    print(str(assignment))
                    for v in variables:
                        while v in pred:
                            pred[pred.index(v)] = assignment[iv]
                        iv += 1
                    fluent = self.unify_fluent_EFP(pred)
                    if 'B(' not in fluent and 'C(' not in fluent and not '-' in fluent:
                        fluents_set.add(fluent)

    def print_precondition_EFP(self,action,out):
        if (len(action.positive_preconditions)+len(action.negative_preconditions) > 0):
            out.write(' if ' )
            #+ str([list(i) for i in action.positive_preconditions]) +  str([list(i) for i in action.negative_preconditions]))
            self.subprint_precondition_EFP(action, 1, out)
            self.subprint_precondition_EFP(action, 0, out)
        out.write(';\n')

    def reorder_bf_list(self, list):
        ret = []
        for elem in list:
            if 'B(' in elem[0]:
                ret.insert(0,elem)
            else:
                ret.append(elem)
        return ret

    def subprint_precondition_EFP(self,action,is_postive,out):
        positive_pre = True
        if (is_postive == 1):
            preconditions = action.positive_preconditions
        else:
            positive_pre = False
            preconditions = action.negative_preconditions
        count = 0
        preconditions = self.reorder_bf_list(preconditions)
        for i in preconditions:
            fluent = self.unify_fluent_EFP(i)
            if (positive_pre):
                out.write(fluent)
            else:
                out.write('-'+ fluent + '')
            if (count < len(preconditions)-1) or (positive_pre and len(action.negative_preconditions) > 0):
                out.write(', ')
                count +=1

    def print_effects_EFP(self,action,out):
        if (action.act_type == 'sensing'):
            act_type = ' determines '
        elif (action.act_type == 'announcement'):
            act_type = ' announces '
        else:
            act_type = ' causes '

        if (len(action.add_effects) > 0):
            for i in action.add_effects:
                out.write(action.name + act_type)
                fluent = self.unify_fluent_EFP(i[0])
                out.write(fluent)
                self.print_conditions_EFP(i[1],i[2],out)

                out.write(';\n')

        if (len(action.del_effects) > 0):
            for i in action.del_effects:
                out.write(action.name + act_type)
                fluent = self.unify_fluent_EFP(i[0])
                out.write('-'+ fluent + '')
                self.print_conditions_EFP(i[1],i[2],out)

                out.write(';\n')

    def print_observers_EFP(self,action,fully,out):
        if fully == 1:
            obs_type = ' observes '
            observers = action.observers
        else:
            obs_type = ' aware_of '
            observers = action.p_observers

        if (len(observers) > 0):
            for ags in observers:
                for ag in ags[0]:
                    if 'FASTART' in ag:
                        for agent in self.objects['agent']:
                            notPrint = 0
                            if ags[3][0][0] != '':
                                if agent == ags[3][0][0]:
                                    notPrint = 1
                            if notPrint == 0:
                                tmp_cond = [[]]
                                self.copy_cond_list(ags,tmp_cond)

                                out.write(agent + obs_type + action.name)
                                self.substitute_ag(tmp_cond[1],agent)
                                self.substitute_ag(tmp_cond[2],agent)

                                self.print_conditions_EFP(tmp_cond[1],tmp_cond[2],out)
                                out.write(';\n')
                    else:
                        out.write(str(ag) + obs_type + action.name)
                        self.print_conditions_EFP(ags[1],ags[2],out)
                        out.write(';\n')

    def copy_cond_list(self, agents, temp):
        i = 0
        while i < len(agents):
            sub_temp = []
            j = 0
            while j < len(agents[i]):
                if i > 0:
                    k = 0
                    sub_sub_temp = []
                    while k < len(agents[i][j]):
                        sub_sub_temp.insert(k,agents[i][j][k])
                        k = k+1
                else:
                    sub_sub_temp = agents[i][j]
                sub_temp.insert(j, sub_sub_temp)
                j = j+1
            temp.insert(i, sub_temp)
            i = i+1

    def substitute_ag(self, conds, agent):
        for cond in conds:
            for elem in cond:
                if 'FASTART' in elem:
                    conds[conds.index(cond)][cond.index(elem)] = re.sub(r'(FASTART\S+FASTOP)', agent ,elem)

    def print_conditions_EFP(self,pos_cond,neg_cond,out):
        yet_to_print = 1
        if self.subprint_cond_EFP(pos_cond,1,out, yet_to_print) == 1:
            yet_to_print = 0;
        self.subprint_cond_EFP(neg_cond,0,out, yet_to_print);

    def subprint_cond_EFP(self,conditions,isPos,out, yet_to_print):
        printed = 0
        for condition in conditions:
            if '' in condition:
                condition.remove('')

        for condition in conditions:
            if not condition:
                conditions.remove(condition)

        if conditions:
            count_cond = 0
            if (yet_to_print == 1):
                out.write( ' if ' )
                printed = 1
            else:
                out.write(', ')
            conditions = self.reorder_bf_list(conditions)
            for condition in conditions:
                cond = self.unify_fluent_EFP(condition)
                if not isPos:
                    out.write('-')
                out.write(cond)
                if count_cond < len(conditions)-1:
                    out.write(', ')
                    count_cond = count_cond +1

        return printed

    #-----------------------------------------------
    # Print PDKB
    #-----------------------------------------------
    def print_PDKB(self,output_folder):
        #########File NAME
        Path(output_folder).mkdir(exist_ok=True)
        return self.print_domain_pdkb(output_folder), self.print_problem_pdkb(output_folder)

    def print_domain_pdkb(self, output_folder):
        pdkb_domain_name = self.domain_name+".pdkbpddl"
        out = open(output_folder + pdkb_domain_name, "w")
        out.write(";This file is automatically generated from an E-PDDL specification and follows the PDKB-PDDL syntax.\n\n")

        out.write(';;;;;;;;;;;;;;;;;;;;    DOMAIN\'S FEATURES    ;;;;;;;;;;;;;;;;;;;\n\n')

        out.write('(define (domain ' + self.domain_name +')\n\n')

        count_types = 0
        out.write('\t(:agents')
        for elem in self.objects['agent']:
            out.write(' ' + elem)
        out.write(')\n\n')

        out.write('\t(:constants)\n\n')

        count_types = 0
        out.write('\t(:types')
        for elem in self.types:
            for el in self.types[elem]:
                out.write('\n\t  ' + el)
                if count_types == len(self.types[elem])-1:
                    out.write('\n\t')
                count_types = count_types+1

        out.write(')\n\n')

        out.write('\t(:predicates')
        for elem in self.predicates:
            out.write( ' (' + elem)
            for el in self.predicates[elem]:
                out.write(' ' + el + ' - ',)
                out.write(self.predicates[elem][el])
            out.write(')')
        out.write(')\n\n')

        out.write(';;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n\n\n')


        out.write(';;;;;;;;;;;;;;;;;    ACTIONS\' SPECIFICATIONS    ;;;;;;;;;;;;;;;;\n\n')

        for action in self.actions:
            out.write(';;;Action ' + action.name + '\n\n')
            out.write('\t(:action ' + action.name + '\n')
            self.print_parameters_PDKB(action, out)
            if not self.print_expl_derive_condition_PDKB(action,out):
                self.print_derive_condition_PDKB(action, out)
            self.print_precondition_PDKB(action, out)
            if not self.print_expl_effects_PDKB(action,out):
                self.print_effects_PDKB(action, out)

            #self.print_observers_EFP(action, 0, out)
            out.write('\t)\n;;;\n\n')
        out.write(';;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n\n\n')

        out.write(')')
        return pdkb_domain_name

    def print_problem_pdkb(self, output_folder):
        pdkb_problem_name = self.problem_name+".pdkbpddl"
        out = open(output_folder + pdkb_problem_name, "w")
        out.write(";This file is automatically generated from an E-PDDL specification and follows the PDKB-PDDL syntax.\n\n")
        out.write('{include:'+self.domain_name+'.pdkbpddl}\n\n')
        out.write('(define (problem '+self.problem_name+')\n')

        out.write('\n\t(:domain ' + self.domain_name + ')\n')
        out.write('\n\t(:depth '+str(self.depth)+')\n')

        if len(self.objects) > 1:
            out.write('\n\t(:objects\n')
            for obj in self.objects:
                if obj != 'agent':
                    out.write('\t\t')
                    for elem in self.objects[obj]:
                        out.write(str(elem)+ ' ')
                    out.write('- ' + str(obj) + '\n')
            out.write('\t)\n')

        out.write('\n\t(:projection )\n')
        out.write('\n\t(:task valid_generation)\n')

        #init print
        out.write('\n\t(:init-type complete)')
        out.write('\n\t(:init')
        t_depth = 1
        while (t_depth <= int(self.depth)):
            out.write("\n\n\t\t;Depth " + str(t_depth))
            for ini_f in self.state:
                ini_fs = self.unify_fluent_init_PDKB(ini_f,t_depth)
                if ini_fs != '':
                    out.write('\n\n' + ini_fs)
            t_depth+=1


        out.write('\n\t)')
        out.write('\n')

        #goal print
        out.write('\n\t(:goal ')
        for goal_f in self.positive_goals:
            goal_fs = self.unify_fluent_PDKB(goal_f)
            out.write('\n\t  ' + goal_fs)

        for goal_f in self.negative_goals:
            goal_fs = self.unify_fluent_PDKB(goal_f)
            out.write('\n\t  !' + goal_fs)
        out.write('\n\t)')
        out.write('\n)')
        return pdkb_problem_name

    def print_parameters_PDKB(self, action, out):
        out.write('\t\t:parameters\t\t\t (')
        count_param = 0
        for param in action.parameters:
            out.write(param[0] + ' - ' + param[1])
            count_param = count_param+1
            if (count_param < len(action.parameters)):
                out.write(' ')
        out.write(')\n')

    def print_expl_derive_condition_PDKB(self,action,out):
        if (len(action.derive_cond) > 0):
            out.write('\t\t:derive-condition\t (')
            out.write(self.unify_fluent_PDKB(action.derive_cond, True))
            out.write(')\n')
            return True

        return False

    def print_derive_condition_PDKB(self,action,out):
        out.write('\t\t:derive-condition\t (')
        observers = action.observers
        visited = False
        if (len(observers) > 0):
            for ags in observers:
                for ag in ags[0]:
                    if 'FASTART' in ag:
                        if visited == True:
                            raise Exception('PDKB coversion cannot handle mutiple Fully Observants Group, make use of the explicit fields.')

                        visited = True
                        if  (len(ags[3]) > 0) and DEBUG:
                            print("\n********CONVERSION WARNING********")
                            print("The \'diff\' operator cannot be directly translated to PDKB and therefore will be ignored.")
                            print("You should make use of the more explicit fields.")
                            print("**********************************\n")

                        if len(ags[1]) == 1 and self.unify_fluent_PDKB(ags[1][0]) != '':

                            if ags[1] == ags[2]:
                                out.write('always')

                            else:
                                der_cond = self.unify_fluent_PDKB(ags[1][0])

                                der_cond = re.sub(r'FASTART(\S)+FASTOP', r'$agent$',der_cond)
                                out.write(der_cond)
                        elif len(ags[2]) == 1:
                            der_cond = self.unify_fluent_PDKB(ags[2][0])
                            der_cond = re.sub(r'FASTART(\S)+FASTOP', r'$agent$',der_cond)
                            out.write('!' + der_cond)
#                            raise Exception('PDKB coversion cannot handle negative condition for Fully Observants Groups, make use of the explicit fields.')
                        elif len(ags[1]) > 1 or len(ags[2]) >1:
                            raise Exception('PDKB coversion cannot handle mutiple Fully Observants Groups, make use of the explicit fields.')
                        else:
                            raise Exception('PDKB coversion cannot handle strange specification for Fully Observants Groups, make use of the explicit fields.')


                        #self.print_conditions_PDKB(ags[1],ags[2],out)
                        out.write(')\n')

        if visited == False:
            out.write('never)\n')

    def print_precondition_PDKB(self,action,out):
        #if (len(action.positive_preconditions)+len(action.negative_preconditions) > 0):
            out.write('\t\t:precondition\t\t (and' )
            #+ str([list(i) for i in action.positive_preconditions]) +  str([list(i) for i in action.negative_preconditions]))
            self.subprint_precondition_PDKB(action, 1, out)
            self.subprint_precondition_PDKB(action, 0, out)
            out.write(')\n')

    def subprint_precondition_PDKB(self,action,is_postive,out):
        positive_pre = True
        if (is_postive == 1):
            preconditions = action.positive_preconditions
        else:
            positive_pre = False
            preconditions = action.negative_preconditions
        for i in preconditions:
            fluent = self.unify_fluent_PDKB(i)

            if not positive_pre:
                fluent = '!'+ fluent

            if not '[' in fluent:
                fluent  = '('+ fluent + ')'

            out.write(' '+ fluent)

    def print_expl_effects_PDKB(self,action,out):
        if (len(action.explicit_eff) > 0):
            out.write('\t\t:effect\t\t\t\t (')
            print ("Effects are: " + str(action.explicit_eff) + " normal are " + str(action.positive_preconditions))
            out.write(self.unify_fluent_PDKB(action.explicit_eff, True))
            out.write(')\n')
            return True

        return False

    def print_effects_PDKB(self,action,out):

    #    if (len(action.p_observers) > 0):
    #        print("\n********CONVERSION WARNING********")
    #        print("Partial observability cannot be directly translated to PDKB and therefore will be ignored.")
    #        print("You should make use of the more explicit fields.")
    #        print("**********************************\n")

        out.write('\t\t:effect\t\t\t\t (and' )
        is_ontic = True;
        if (action.act_type == 'sensing' or action.act_type == 'announcement'):
            is_ontic = False

        printed = False
        self.subprint_effects_PDKB(action, out, is_ontic, printed, True)
        self.subprint_effects_PDKB(action, out, is_ontic, printed, False)
        out.write(')\n')

    def subprint_effects_PDKB(self,action,out,is_ontic,printed, is_pos):
        ag_printed = False
        p_ag_printed = False

        count = 0
        if (is_pos):
            effects = action.add_effects
        else:
            effects = action.del_effects

        if (len(effects) > 0):
            for i in effects:
                if count == 3:
                    count = 0
                    out.write('\n\t\t\t\t\t\t\t ')
                fluent = self.unify_fluent_PDKB(i[0])
                if self.print_conditions_PDKB(i[1],i[2],out) == 1:
                    printed = True
                if (is_ontic):
                    count = count + 1
                    if (is_pos):
                        out.write(' ('+ fluent + ')')
                    else:
                        out.write(' (!'+ fluent + ')')


                if (len(action.observers) > 0):
                    for ags in action.observers:
                        for ag in ags[0]:
                            if not 'FASTART' in ag:
                                if self.print_conditions_PDKB(ags[1],ags[2],out) == 1:
                                    ag_printed = True
                                else:
                                    ag_printed = False
                                if count == 3:
                                    count = 0
                                    out.write('\n\t\t\t\t\t\t\t ')
                                    count = count + 1
                                out.write(' [' + ag + '](')
                                if (is_pos):
                                    out.write(fluent + ')')
                                else:
                                    out.write('!'+fluent + ')')

                                if (len(action.p_observers) > 0):
                                    for p_ags in action.p_observers:
                                        for p_ag in p_ags[0]:
                                            if not 'FASTART' in p_ag:
                                                if self.print_conditions_PDKB(p_ags[1],p_ags[2],out) == 1:
                                                    p_ag_printed = True
                                                else:
                                                    p_ag_printed = False
                                                if count == 3:
                                                    count = 0
                                                    out.write('\n\t\t\t\t\t\t\t ')
                                                count = count + 1
                                                out.write(' [' + p_ag + '][' + ag + '](or')
                                                out.write(' ('+fluent + ')')
                                                out.write(' (!'+fluent + '))')
                                            elif (DEBUG):
                                                print("\n********CONVERSION WARNING********")
                                                print("Partial observability has not fully integrated the FORALL operator.")
                                                print("You should make use of the more explicit fields if the results are not correct.")
                                                print("**********************************\n")
                                if p_ag_printed:
                                    out.write(')')
                                    p_ag_printed = False

                            elif (DEBUG):
                                print("\n********CONVERSION WARNING********")
                                print("Observability has not fully integrated the FORALL operator.")
                                print("Please, check the resulting conversion.\nYou should make use of the more explicit fields if the results are not correct.")
                                print("**********************************\n")
                    if ag_printed:
                        out.write(')')
                        ag_printed = False

                        #print ("Obs of " + action.name + ": " + str(ags))
                        #for ag in ags[0]:
                        #    if not 'FASTART' in ag:
                        #        if count == 3:
                        #            count = 0
                        #            out.write('\n\t\t\t\t\t\t\t ')
                        #        count = count + 1
                        #        out.write(' [' + ag + '](')
                        #        if (is_pos):
                        #            out.write(fluent + ')')
                        #        else:
                        #            out.write('!'+fluent + ')')



                elif not is_ontic:
                    raise Exception('Each action needs at least one fully observant agent.')
                if printed == True:
                    out.write(')')

    def print_conditions_PDKB(self,pos_cond,neg_cond,out):
        yet_to_print = 1
        if self.subprint_cond_PDKB(pos_cond,1,out, yet_to_print) == 1:
            yet_to_print = 0;
        if self.subprint_cond_PDKB(neg_cond,0,out, yet_to_print) == 1 or yet_to_print == 0:
            out.write(")")
            return 1
        return 0

    def subprint_cond_PDKB(self,conditions,isPos,out, yet_to_print):
        printed = 0
        for condition in conditions:
            if '' in condition:
                condition.remove('')

        for condition in conditions:
            if not condition:
                conditions.remove(condition)

        if conditions:
            count_cond = 0
            if (yet_to_print == 1):
                out.write( ' (when (and (' )
                printed = 1
            else:
                out.write(' (')
            for condition in conditions:
                cond = self.unify_fluent_PDKB(condition)
                if not isPos:
                    out.write('!')
                out.write(cond)
                if count_cond < len(conditions):
                    out.write(')')
                    count_cond = count_cond +1
        return printed

    def unify_fluent_init_PDKB(self,given_list, t_depth, no_change = False):
        ret = ''
        count = 1
        found = False
        to_reprint = False

        if t_depth == 1:
            to_reprint = True

        new_list = copy.deepcopy(given_list)
        for elem in given_list:
            if 'C(' in elem:
                if len(self.objects['agent']) == elem.count(','):
                    to_reprint = True
                    if not found:
                        found = True
                    else:
                        return ('\t\t('+ self.unify_fluent_PDKB(given_list)+') ')

                    del new_list[new_list.index(elem)]

                    while count <= t_depth:
                        ag_name = '?ag'+str(count)
                        new_list.insert(0,'B('+ag_name+',')
                        t_count = 0
                        ret += ('\t\t(forall ' + ag_name + ' - agent\n')
                        while t_count < count:
                            ret += '\t'
                            t_count += 1
                        count += 1

        if to_reprint:
            ret+='\t\t'
            if not found:
                ret+= '('
                count += 1

            ret += self.unify_fluent_PDKB(new_list)

            while count > 1:
                ret += ')'
                count -= 1
            #ret+='\n'

        return ret

    def unify_fluent_PDKB(self,given_list, no_change = False):
        return Action.unify_fluent_PDKB(given_list, no_change, False)


#-----------------------------------------------
# Main
#-----------------------------------------------
if __name__ == '__main__':
    import sys, pprint
    domain = sys.argv[1]
    problem = sys.argv[2]
    parser = EPDDL_Parser()
#    print('----------------------------')
#    pprint.pprint(parser.scan_tokens(domain))
#    print('----------------------------')
#    pprint.pprint(parser.scan_tokens(problem))
#    print('----------------------------')
    parser.parse_domain(domain)
    parser.parse_problem(problem)
    output_folder = "exp/tmp"
    parser.print_EFP(output_folder)
    print("\nThe given files have been correctly converted to mAp.")
    print("The resulting file, called \'" +parser.domain_name+"_"+parser.problem_name+".tmp\', is in the \'"+output_folder+"\' folder.\n")


    f = open("exp/tmp/lastcreated.tmp", "w")

    p, domain_name_from_file = os.path.split(domain)
    domain_name_from_file = domain_name_from_file.rsplit('.', 1)[0]

    p, problem_name_from_file = os.path.split(problem)
    problem_name_from_file = problem_name_from_file.rsplit('.', 1)[0]

    f.write(domain_name_from_file+"_"+problem_name_from_file+".tmp")
    f.close()


    #parser.print_PDKB()
    #print("\nThe given files have been correctly converted to PDKB-PDDL.")
    #print("The resulting files, called \'" +parser.domain_name+".pdkpddl\' and \'" +parser.problem_name+".pdkpddl\', are in the \'out\pdkb\' folder.\n")
#    print('State: ' + str(parser.state))
#    for act in parser.actions:
#        print(act)
#    for action in parser.actions:
#        for act in action.groundify(parser.objects, parser.types, parser.requirements, fluents):
#            print(act)
#    print('----------------------------')
#    print('Problem name: ' + parser.problem_name)
#    print('Objects: ' + str(parser.objects))
    #print('Predicates: ' + str(parser.predicates)
#    print('State: ' + str(parser.state))
#    print('Positive goals: ' + str(parser.positive_goals))
#    print('Negative goals: ' + str(parser.negative_goals))
